# coding=utf-8
# Copyright 2023 The Uncertainty Baselines Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ensemble on CIFAR.

This script only performs evaluation, not training. We recommend training
ensembles by launching independent runs of `deterministic.py` over different
seeds.
"""

import os

from absl import app
from absl import flags
from absl import logging

import numpy as np
import robustness_metrics as rm
import tensorflow as tf
import tensorflow_datasets as tfds
import uncertainty_baselines as ub
import ood_utils  # local file import from baselines.cifar
import utils  # local file import from baselines.cifar
import pickle
os.environ["CUDA_VISIBLE_DEVICES"] = "1"

# flags.DEFINE_string('checkpoint_dir', None,
#                     'The directory where the model weights are stored.')
# flags.mark_flag_as_required('checkpoint_dir')

# OOD flags.
flags.DEFINE_bool('eval_on_ood', False,
									'Whether to run OOD evaluation on specified OOD datasets.')
flags.DEFINE_list('ood_dataset', 'cifar100,svhn_cropped',
									'list of OOD datasets to evaluate on.')
flags.DEFINE_bool('dempster_shafer_ood', False,
									'Wheter to use DempsterShafer Uncertainty score.')

flags.DEFINE_string('corruption_type', None, 'corruption_type')
flags.DEFINE_integer('severity', None, 'Number')
FLAGS = flags.FLAGS
# FLAGS.dataset = "cifar100"

def parse_checkpoint_dir(checkpoint_dir):
	"""Parse directory of checkpoints."""
	paths = []
	subdirectories = tf.io.gfile.glob(os.path.join(checkpoint_dir, '*'))
	is_checkpoint = lambda f: ('checkpoint' in f and '.index' in f)
	for subdir in subdirectories:
		for path, _, files in tf.io.gfile.walk(subdir):
			if any(f for f in files if is_checkpoint(f)):
				latest_checkpoint_without_suffix = tf.train.latest_checkpoint(path)
				paths.append(os.path.join(path, latest_checkpoint_without_suffix))
				break
	return paths


def main(argv):
	del argv  # unused arg
	if not FLAGS.use_gpu:
		raise ValueError('Only GPU is currently supported.')
	if FLAGS.num_cores > 1:
		raise ValueError('Only a single accelerator is currently supported.')
	tf.random.set_seed(FLAGS.seed)
	tf.io.gfile.makedirs(FLAGS.output_dir)
	logging.info('output_dir=%s', FLAGS.output_dir)

	ds_info = tfds.builder(FLAGS.dataset).info
	batch_size = FLAGS.per_core_batch_size * FLAGS.num_cores
	num_classes = ds_info.features['label'].num_classes

	data_dir = FLAGS.data_dir

	# dataset_builder = ub.datasets.get(
	# 		f'{FLAGS.dataset}_1',
	# 		corruption_type="v6",
	# 		# downloaddata=True,
	# 		split=tfds.Split.TEST,
	# 		data_dir=data_dir,
	# 		drop_remainder=FLAGS.drop_remainder_for_eval)

	# dataset_builder = ub.datasets.get(
	# 		f'{FLAGS.dataset}_corrupted',
	# 		corruption_type=FLAGS.corruption_type,
	# 		# downloaddata=True,
	# 		severity=FLAGS.severity,
	# 		split=tfds.Split.TEST,
	# 		data_dir=data_dir + "/cifar100_corrupted",
	# 		drop_remainder=FLAGS.drop_remainder_for_eval)

	dataset_builder = ub.datasets.get(
			"cifar100",
			download_data=FLAGS.download_data,
			data_dir=data_dir,
			split=tfds.Split.TEST,
			drop_remainder=FLAGS.drop_remainder_for_eval)
	
	dataset = dataset_builder.load(batch_size=batch_size)
	steps_per_eval = 10000 // batch_size
	test_datasets = {
		'clean': dataset
		}
	# test_datasets[f'{FLAGS.corruption_type}_{FLAGS.severity}'] = (dataset)
	# test_datasets = {}
	if FLAGS.eval_on_ood:
		ood_dataset_names = FLAGS.ood_dataset
		ood_datasets, steps_per_ood = ood_utils.load_ood_datasets(
				ood_dataset_names,
				dataset_builder,
				1. - FLAGS.train_proportion,
				batch_size,
				drop_remainder=FLAGS.drop_remainder_for_eval)
		test_datasets.update(ood_datasets)
	# corruption_types, _ = utils.load_corrupted_test_info(FLAGS.dataset)
	# for corruption_type in corruption_types:
	# 	for severity in range(1, 6):
	# 		dataset = ub.datasets.get(
	# 				f'{FLAGS.dataset}_corrupted',
	# 				corruption_type=corruption_type,
	# 				download_data=FLAGS.download_data,
	# 				data_dir=data_dir,
	# 				severity=severity,
	# 				split=tfds.Split.TEST,
	# 				drop_remainder=FLAGS.drop_remainder_for_eval).load(
	# 						batch_size=batch_size)
	# 		if corruption_type == "frosted_glass_blur" and severity == 5:
	# 			dataset_builder = ub.datasets.get(
	# 				f'{FLAGS.dataset}_corrupted',
	# 				corruption_type="frosted_glass_blur",
	# 				severity = 5,
	# 				# downloaddata=True,
	# 				split=tfds.Split.TEST,
	# 				data_dir=data_dir,
	# 				drop_remainder=FLAGS.drop_remainder_for_eval)

	# 			# dataset_builder = ub.datasets.get(
	# 			# 	f'{FLAGS.dataset}_1',
	# 			# 	corruption_type="v6",
	# 			# 	# downloaddata=True,
	# 			# 	split=tfds.Split.TEST,
	# 			# 	data_dir=data_dir,
	# 			# 	drop_remainder=FLAGS.drop_remainder_for_eval)
				
	# 			steps_per_eval = dataset_builder.num_examples // batch_size
	# 			dataset = dataset_builder.load(batch_size=batch_size)
	# 			test_datasets[f'{corruption_type}_{severity}'] = dataset

	model = ub.models.wide_resnet(
			input_shape=ds_info.features['image'].shape,
			depth=28,
			width_multiplier=10,
			num_classes=num_classes,
			l2=0.,
			version=2)
	logging.info('Model input shape: %s', model.input_shape)
	logging.info('Model output shape: %s', model.output_shape)
	logging.info('Model number of weights: %s', model.count_params())

	# Search for checkpoints from their index file; then remove the index suffix.
	ensemble_filenames = ["utils/out/deterministic1_cifar10/checkpoint-9", "utils/out/deterministic2_cifar10/checkpoint-9",
		"utils/out/deterministic3_cifar10/checkpoint-9", "utils/out/deterministic4_cifar10/checkpoint-9"] 
	ensemble_size = len(ensemble_filenames)
	logging.info('Ensemble size: %s', ensemble_size)
	logging.info('Ensemble number of weights: %s',
							 ensemble_size * model.count_params())
	logging.info('Ensemble filenames: %s', str(ensemble_filenames))
	checkpoint = tf.train.Checkpoint(model=model)
	# Write model predictions to files.
	num_datasets = len(test_datasets)
	for m, ensemble_filename in enumerate(ensemble_filenames):
		checkpoint.restore(ensemble_filename)
		for n, (name, test_dataset) in enumerate(test_datasets.items()):
			filename = '{dataset}_{member}.npy'.format(
					dataset=name.replace('/', '_'), member=m)  # ood dataset name has '/'
			filename = os.path.join(FLAGS.output_dir, filename)
			if not tf.io.gfile.exists(filename):
				logits = []
				test_iterator = iter(test_dataset)
				steps = steps_per_eval if 'ood/' not in name else steps_per_ood[name]
				for _ in range(steps):
					features = next(test_iterator)['features']  # pytype: disable=unsupported-operands
					logits.append(model(features, training=False))

				logits = tf.concat(logits, axis=0)
				with tf.io.gfile.GFile(filename, 'w') as f:
					np.save(f, logits.numpy())
			percent = (m * num_datasets + (n + 1)) / (ensemble_size * num_datasets)
			message = ('{:.1%} completion for prediction: ensemble member {:d}/{:d}. '
								 'Dataset {:d}/{:d}'.format(percent,
																						m + 1,
																						ensemble_size,
																						n + 1,
																						num_datasets))
			logging.info(message)

	metrics = {
			'test/negative_log_likelihood': tf.keras.metrics.Mean(),
			'test/gibbs_cross_entropy': tf.keras.metrics.Mean(),
			'test/accuracy': tf.keras.metrics.SparseCategoricalAccuracy(),
			'test/ece': rm.metrics.ExpectedCalibrationError(
					num_bins=FLAGS.num_bins),
			'test/diversity': rm.metrics.AveragePairwiseDiversity(),
	}
	if FLAGS.eval_on_ood:
		ood_metrics = ood_utils.create_ood_metrics(ood_dataset_names)
		metrics.update(ood_metrics)
	corrupt_metrics = {}
	for name in test_datasets:
		corrupt_metrics['test/nll_{}'.format(name)] = tf.keras.metrics.Mean()
		corrupt_metrics['test/accuracy_{}'.format(name)] = (
				tf.keras.metrics.SparseCategoricalAccuracy())
		corrupt_metrics['test/ece_{}'.format(name)] = (
				rm.metrics.ExpectedCalibrationError(num_bins=FLAGS.num_bins))
	for i in range(ensemble_size):
		metrics['test/nll_member_{}'.format(i)] = tf.keras.metrics.Mean()
		metrics['test/accuracy_member_{}'.format(i)] = (
				tf.keras.metrics.SparseCategoricalAccuracy())

	# Evaluate model predictions.
	list_probs = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
	list_labels = tf.TensorArray(tf.int32, size=0, dynamic_size=True)
	
	for n, (name, test_dataset) in enumerate(test_datasets.items()):
		logits_dataset = []
		for m in range(ensemble_size):
			filename = '{dataset}_{member}.npy'.format(
					dataset=name.replace('/', '_'), member=m)  # ood dataset name has '/'
			filename = os.path.join(FLAGS.output_dir, filename)
			with tf.io.gfile.GFile(filename, 'rb') as f:
				logits_dataset.append(np.load(f))

		logits_dataset = tf.convert_to_tensor(logits_dataset)
		test_iterator = iter(test_dataset)
		steps = steps_per_eval if 'ood/' not in name else steps_per_ood[name]
		for step in range(steps):
			inputs = next(test_iterator)
			labels = inputs['labels']  # pytype: disable=unsupported-operands
			logits = logits_dataset[:, (step*batch_size):((step+1)*batch_size)]
			labels = tf.cast(labels, tf.int32)
			negative_log_likelihood_metric = rm.metrics.EnsembleCrossEntropy()
			negative_log_likelihood_metric.add_batch(logits, labels=labels)
			negative_log_likelihood = list(
					negative_log_likelihood_metric.result().values())[0]
			per_probs = tf.nn.softmax(logits)
			probs = tf.reduce_mean(per_probs, axis=0)
			list_probs = list_probs.write(list_probs.size(), probs)

			logits_mean = tf.reduce_mean(logits, axis=0)
			list_labels = list_labels.write(list_labels.size(), labels)

			if name == 'clean':
			  gibbs_ce_metric = rm.metrics.GibbsCrossEntropy()
			  gibbs_ce_metric.add_batch(logits, labels=labels)
			  gibbs_ce = list(gibbs_ce_metric.result().values())[0]
			  metrics['test/negative_log_likelihood'].update_state(
			      negative_log_likelihood)
			  metrics['test/gibbs_cross_entropy'].update_state(gibbs_ce)
			  metrics['test/accuracy'].update_state(labels, probs)
			  metrics['test/ece'].add_batch(probs, label=labels)

			  for i in range(ensemble_size):
			    member_probs = per_probs[i]
			    member_loss = tf.keras.losses.sparse_categorical_crossentropy(
			        labels, member_probs)
			    metrics['test/nll_member_{}'.format(i)].update_state(member_loss)
			    metrics['test/accuracy_member_{}'.format(i)].update_state(
			        labels, member_probs)
			  metrics['test/diversity'].add_batch(per_probs)

			elif name.startswith('ood/'):
			  ood_labels = 1 - inputs['is_in_distribution']  # pytype: disable=unsupported-operands
			  if FLAGS.dempster_shafer_ood:
			    ood_scores = ood_utils.DempsterShaferUncertainty(logits_mean)
			  else:
			    ood_scores = 1 - tf.reduce_max(probs, axis=-1)

			  for metric_name, metric in metrics.items():
			    if name in metric_name:
			      metric.update_state(ood_labels, ood_scores)
			else:
			  corrupt_metrics['test/nll_{}'.format(name)].update_state(
			      negative_log_likelihood)
			  corrupt_metrics['test/accuracy_{}'.format(name)].update_state(
			      labels, probs)
			  corrupt_metrics['test/ece_{}'.format(name)].add_batch(
			      probs, label=labels)

		message = ('{:.1%} completion for evaluation: dataset {:d}/{:d}'.format(
		    (n + 1) / num_datasets, n + 1, num_datasets))
		logging.info(message)
	list_probs = list_probs.stack()
	list_labels = list_labels.stack()
	list_probs = list_probs.numpy()
	list_probs = list_probs.reshape((list_probs.shape[0] * list_probs.shape[1]), list_probs.shape[2])
	list_labels = list_labels.numpy()
	# ood_type = f'{FLAGS.corruption_type}_{FLAGS.severity}'
	ood_type = "clean"
	with open("out/ood_cifar/deepensembles/" + ood_type + "_probs.pkl", "wb") as fp:
		pickle.dump(list_probs, fp)
	with open("out/ood_cifar/deepensembles/" + ood_type + "_labels.pkl", "wb") as fp:
		pickle.dump(list_labels.ravel(), fp)
	
	# logging.info('Test NLL: %.4f, Accuracy: %.2f%%, ECE: %.4f',
	# 				metrics['test/negative_log_likelihood'].result(),
	# 				metrics['test/accuracy'].result() * 100,
	# 				metrics['test/ece'].result()['ece'],)

	# logging.info('Test NLL: %.4f, Accuracy: %.2f%%, ECE: %.4f',
	# 				corrupt_metrics['test/nll_{}'.format(ood_type)].result(),
	# 				corrupt_metrics['test/accuracy_{}'.format(ood_type)].result() * 100,
	# 				corrupt_metrics['test/ece_{}'.format(ood_type)].result()['ece'],)
	exit()
	
	corrupt_results = utils.aggregate_corrupt_metrics(corrupt_metrics,
																										corruption_types)
	total_results = {name: metric.result() for name, metric in metrics.items()}
	total_results.update(corrupt_results)
	# Results from Robustness Metrics themselves return a dict, so flatten them.
	total_results = utils.flatten_dictionary(total_results)
	logging.info('Metrics: %s', total_results)
	logging.info('Test NLL: %.4f, Accuracy: %.2f%%, ECE: %.4f',
					metrics['test/negative_log_likelihood'].result(),
					metrics['test/accuracy'].result() * 100,
					metrics['test/ece'].result()['ece'],)

if __name__ == '__main__':
	app.run(main)
